var rightLinks = {
	pu: rightLinksPrefUtils,

	enabled: true,
	stopContextMenu: false,
	runned: false,
	item: null,
	origItem: null,
	itemType: null,
	event: null,
	XULNS: "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
	showContextTimeout: null,

	init: function() {
		window.removeEventListener("load", this, false);
		window.addEventListener("unload", this, false);

		this.setListeners(["mousedown", "mouseup", "click", "contextmenu", "popupshowing"], true);
		this.pu.init();
		this.setStatus();
		this.setItemsMode();
		this.registerHotkeys();
	},
	destroy: function() {
		window.removeEventListener("unload", this, false);
		this.setListeners(["mousedown", "mouseup", "click", "contextmenu", "popupshowing"], false);
		this.cancelDelayedContext();
		this.pu.destroy();
	},
	handleEvent: function(e) {
		switch(e.type) {
			case "load":         this.init(e);                break;
			case "unload":       this.destroy(e);             break;
			case "mousedown":    this.mousedownHandler(e);    break;
			case "mouseup":      this.mouseupHandler(e);      break;
			case "click":        this.clickHandler(e);        break;
			case "contextmenu":  this.contextmenuHandler(e);  break;
			case "popupshowing": this.popupshowingHandler(e); break;
			case "mousemove":    this.mousemoveHandler(e);
		}
	},
	setListeners: function(evtTypes, addFlag) {
		var act = addFlag ? "addEventListener" : "removeEventListener";
		evtTypes.forEach(function(evtType) {
			window[act](evtType, this, true);
		}, this);
	},

	isVoidURI: function(uri) {
		uri = (uri || "").replace(/(?:\s|%20)+/g, " ");
		return /^javascript: *(?:|\/\/|void *(?: +0|\( *0 *\))) *;? *$/i.test(uri);
	},
	isJSURI: function(uri) {
		return typeof uri == "string" && /^javascript:/i.test(uri);
	},
	isDummyURI: function(item, uri, evt) {
		//if(this.itemType != "link")
		//	return false;
		item = item || this.item;
		uri = uri || this.getHref(item, evt || this.event);
		var doc = item.ownerDocument;
		var loc = doc.location.href.replace(/#.*$/, "");
		if(!this.hasPrefix(uri, loc))
			return false;
		var _uri = uri.substr(loc.length);
		if(_uri == "" && item.getAttribute && item.hasAttribute("href") && !item.getAttribute("href")) // <a href="">
			return true;
		if(_uri.charAt(0) != "#")
			return false;
		var anchor = _uri.substr(1);
		if(anchor.charAt(0) == "!") // site.com/#!... links on JavaScript-based sites like http://twitter.com/
			return false;
		return !anchor || !doc.getElementById(anchor) && !doc.getElementsByName(anchor).length;
	},
	get mi() {
		return document.getElementById("rightLinks-toggleStatus-mi");
	},
	get miApp() {
		return document.getElementById("rightLinks-toggleStatus-mi-app");
	},
	get miAppSep() {
		return document.getElementById("rightLinks-toggleStatus-mi-app-separator");
	},
	get status() {
		return document.getElementById("rightLinks-toggleStatus-status");
	},
	get tbb() {
		return document.getElementById("rightLinks-toggleStatus-tbutton");
	},
	get paletteButton() {
		var tb = "gNavToolbox" in window && gNavToolbox
			|| "getNavToolbox" in window && getNavToolbox() // Firefox 3.0
			|| document.getElementById("navigator-toolbox"); // Firefox <= 2.0
		if(!tb || !("palette" in tb))
			return null;
		var elt = tb.palette.getElementsByAttribute("id", "rightLinks-toggleStatus-tbutton");
		return elt.length ? elt[0] : null;
	},
	get appInfo() {
		delete this.appInfo;
		return this.appInfo = Components.classes["@mozilla.org/xre/app-info;1"]
			.getService(Components.interfaces.nsIXULAppInfo);
	},
	get isSeaMonkey() {
		delete this.isSeaMonkey;
		return this.isSeaMonkey = this.appInfo.name == "SeaMonkey";
	},
	get fxVersion() {
		var ver = parseFloat(this.appInfo.version); // 3.0 for "3.0.10"
		if(this.isSeaMonkey) switch(ver) {
			case 2:            ver = 3.5; break;
			case 2.1: default: ver = 4;
		}
		delete this.fxVersion;
		return this.fxVersion = ver;
	},
	getItem: function(e) {
		if(e.button != 2 || e.ctrlKey || e.shiftKey || e.altKey || e.metaKey || !this.enabled)
			return null;
		var it = e.originalTarget;
		if(!it.localName) // it === document
			return null;
		var _it;

		_it = this.getLink(it);
		if(_it) {
			this.itemType = "link";
			return _it;
		}

		if(this.pu.pref("enabledOnHistoryItems")) {
			_it = this.getHistoryItem(it, e);
			if(_it) {
				this.itemType = "historyItem";
				return _it;
			}
		}
		if(this.pu.pref("enabledOnBookmarks")) {
			_it = this.getBookmarkItem(it, e);
			if(_it) {
				this.itemType = "bookmark";
				return _it;
			}
		}
		if(this.pu.pref("enabledOnImages")) {
			_it = this.getImg(it);
			if(
				_it
				&& (
					this.pu.pref("enabledOnSingleImages")
					|| _it.ownerDocument.documentURI != _it.src
				)
			) {
				this.itemType = "img";
				return _it;
			}
		}
		return null;
	},

	getLink: function(it) {
		if(!it || !it.localName)
			return null;
		if(
			it.namespaceURI == this.XULNS
			&& this.inObject(it, "href") && (it.href || it.hasAttribute("href"))
			&& (it.accessibleType || it.wrappedJSObject.accessibleType) == Components.interfaces.nsIAccessibleProvider.XULLink
		)
			return it;

		const docNode = Node.DOCUMENT_NODE; // 9
		const eltNode = Node.ELEMENT_NODE; // 1
		for(it = it; it && it.nodeType != docNode; it = it.parentNode) {
			// https://bugzilla.mozilla.org/show_bug.cgi?id=266932
			// https://bug266932.bugzilla.mozilla.org/attachment.cgi?id=206815
			// It's strange to see another link in Status Bar
			// and other browsers (Opera, Safari, Google Chrome) will open "top level" link.
			// And IE... IE won't open XML (it's important!) testcase. :D
			// Also this seems like bug of left-click handler.
			// So, let's open link, which user see in Status Bar.
			if(
				(
					it instanceof HTMLAnchorElement
					|| it instanceof HTMLAreaElement
					|| it instanceof HTMLLinkElement
				)
				&& (
					it.hasAttribute("href")
					|| this.getProperty(it, "repObject", "href") // Firebug
				)
				|| it.nodeType == eltNode && it.hasAttributeNS("http://www.w3.org/1999/xlink", "href")
			)
				return it;
		}
		return null;
	},
	getHistoryItem: function(it, e) {
		if(!it || !it.localName)
			return null;
		if(it.namespaceURI != this.XULNS)
			return null;
		var itln = it.localName.toLowerCase();
		if(
			(
				this.hasParent(it, "goPopup")
				|| itln == "treechildren" && (it.parentNode.id || "").toLowerCase().indexOf("history") != -1 // Sidebar
			)
			&& this.getBookmarkURI(it, e)
		)
			return it;
		return null;
	},
	getBookmarkItem: function(it, e) {
		if(!it || !it.localName)
			return null;
		if(it.namespaceURI != this.XULNS)
			return null;
		var itln = it.localName.toLowerCase();
		if(
			!("type" in it && it.type == "menu")
			&& (
				(
					/(?:^|\s)bookmark-item(?:\s|$)/.test(it.className)
					&& (itln == "toolbarbutton" || itln == "menuitem")
				)
				|| itln == "menuitem" && it.hasAttribute("siteURI")
				|| itln == "treechildren" && (it.parentNode.id || "").toLowerCase().indexOf("bookmark") != -1 // Sidebar
			)
			&& !this.hasParent(it, "goPopup")
			&& this.getBookmarkURI(it, e)
		)
			return it;
		return null;
	},
	getImg: function(it) {
		if(!it || !it.localName)
			return null;
		var itln = it.localName.toLowerCase();
		if(itln == "_moz_generated_content_before") { // Alt-text
			it = it.parentNode;
			itln = it.localName.toLowerCase();
		}
		if(
			(
				(itln == "img" || itln == "image") && it.hasAttribute("src")
				|| it instanceof HTMLCanvasElement
			)
			&& !this.isChromeDoc(it.ownerDocument) // Not for interface...
			&& ( // Speed Dial has own settings for right clicks
				it.ownerDocument.documentURI != "chrome://speeddial/content/speeddial.xul"
				|| !/(?:^|\s)speeddial-container(?:\s|$)/.test(it.parentNode.className)
				|| this.pu.pref("enabledOnSpeedDialImages")
			)
		)
			return it;
		return null;
	},

	hasParent: function(it, pId) {
		for(it = it.parentNode; it && "id" in it; it = it.parentNode)
			if(it.id == pId)
				return true;
		return false;
	},
	getBookmarkURI:	function(it, e, getPlacesURIs) {
		var ln = it.localName;
		var uri = ln && ln.toLowerCase() == "treechildren"
			? this.getTreeInfo(it, e, "uri")
			: it.statusText
				|| it.node && it.node.uri
				|| it._placesNode && it._placesNode.uri // Firefox 3.7a5pre
				|| it.getAttribute("siteURI")
				|| "";
		return !getPlacesURIs && /^place:/.test(uri) ? "" : uri;
	},
	getTreeInfo: function(treechildren, e, prop) { // "uri" or "title"
		if(!("PlacesUtils" in window)) // For Firefox 3.0+
			return "";
		var tree = (treechildren || this.item).parentNode
		e = e || this.event;

		// Based on code of Places' Tooltips ( https://addons.mozilla.org/firefox/addon/7314 )
		var row = {}, column = {}, part = {};
		var tbo = tree.treeBoxObject;
		tbo.getCellAt(e.clientX, e.clientY, row, column, part);
		if(row.value == -1)
			return "";
		var node = tree.view.nodeForTreeIndex(row.value);
		if(!PlacesUtils.nodeIsURI(node))
			return "";
		return node[prop];
	},
	getLinkURI: function(it) {
		const ns = "http://www.w3.org/1999/xlink";
		return it.hasAttributeNS(ns, "href")
			? makeURLAbsolute(it.baseURI, it.getAttributeNS(ns, "href")) // See chrome://browser/content/utilityOverlay.js
			// Looks like wrapper error with chrome://global/content/bindings/text.xml#text-link binding
			// on "content" pages (e.g. chrome://global/content/console.xul)
			: it.href || it.getAttribute("href")
				|| this.getProperty(it, "repObject", "href"); // Firebug
	},
	getHref: function(a, e) {
		a = a || this.item;
		e = e || this.event;
		return this.getLinkURI(a)
			|| a.src || a.getAttribute("src")
			|| a instanceof HTMLCanvasElement && a.toDataURL()
			|| this.getBookmarkURI(a, e, "uri");
	},
	closeMenus: function(it) {
		it = it || this.item;
		if(it && typeof it == "object")
			closeMenus(it); // chrome://browser/content/utilityOverlay.js
	},
	stopEvent: function(e) {
		e.preventDefault();
		e.stopPropagation();
	},
	isPrimitive: function(v) {
		if(v === null || v === undefined)
			return true;
		var t = typeof v;
		return t == "string" || t == "number" || t == "boolean";
	},
	getProperty: function(obj) {
		var u;
		if(this.isPrimitive(obj))
			return u;
		var a = arguments, p;
		for(var i = 1, len = a.length - 1; i <= len; i++) {
			p = a[i];
			if(!(p in obj))
				return u;
			obj = obj[p];
			if(i == len)
				return obj;
			if(this.isPrimitive(obj))
				return u;
		}
		return u;
	},
	isChromeDoc: function(doc) {
		return doc.defaultView instanceof Components.interfaces.nsIDOMChromeWindow;
	},
	inObject: function(o, p) {
		// this._log("inObject", "wrappedJSObject" in o, o.wrappedJSObject);
		// Open chrome://global/content/console.xul in tab
		// and click on <xul:label class="text-link" />
		//   "wrappedJSObject" in o => false
		//   o.wrappedJSObject      => [object XULElement]

		//return p in o || "wrappedJSObject" in o && p in o.wrappedJSObject;
		return p in o || o.wrappedJSObject && p in o.wrappedJSObject;
	},
	hasPrefix: function(str, prefix) {
		return str.substr(0, prefix.length) == prefix;
	},

	mousedownHandler: function(e) {
		if(!this.enabled)
			return;

		if("_cleanupTimer" in this) {
			clearTimeout(this._cleanupTimer);
			delete this._cleanupTimer;
		}

		this.item = this.getItem(e);
		this.origItem = e.originalTarget;
		this.runned = false;
		this.stopContextMenu = !!this.item;
		if(!this.item)
			return;
		var clone = {};
		for(var p in e) // Firefox 1.5.0.x and 2.0.0.x bug
			clone[p] = e[p];
		this.event = clone;
		this.saveXY(e);
		var cmt = this.pu.pref("showContextMenuTimeout");
		if(cmt > 0) {
			this.cancelDelayedContext();
			this.showContextTimeout = setTimeout(
				function(_this) {
					_this.runned = true;
					_this.showContextMenu();
				},
				cmt,
				this
			);
		}

		this.removeMoveHandler();
		this.disallowMousemove = this.pu.pref("disallowMousemoveDist") >= 0;
		this.mousemoveParams = {
			dist: 0,
			screenX: e.screenX,
			screenY: e.screenY
		};
		window.addEventListener("mousemove", this, true);
	},
	mouseupHandler: function(e) {
		if(!this.enabled)
			return;
		this.saveXY(e);
		this.removeMoveHandler();
		/*
		setTimeout(function(_this) {
			_this.runned = false;
			_this.stopContextMenu = false;
			_this.removeMoveHandler();
		}, 0, this);
		*/
		setTimeout(function(_this) {
			_this.stopContextMenu = false;
		}, 0, this);
		this._cleanupTimer = setTimeout(function(_this) {
			delete this._cleanupTimer;
			_this.item = _this.origItem = _this.event = null;
		}, 500, this);
	},
	clickHandler: function(e) {
		if(!this.enabled || this.runned)
			return;

		var a = this.getItem(e);
		if(!a)
			return;
		var href = this.getHref(a, e);
		if(!href)
			return;
		this.stopEvent(e);
		this.cancelDelayedContext();
		this.item = a;
		this.origItem = e.originalTarget;
		if((this.itemType == "bookmark" || this.itemType == "historyItem") && this.pu.pref("closePopups"))
			this.closeMenus(a);

		var voidURI = this.isVoidURI(href);
		if(!voidURI && this.isJSURI(href)) {
			var _this = this;
			var loadJSFunc = function() { _this.loadJSLink(); };
			var loadJS = this.pu.pref("loadJavaScriptLinks");
			if(this.pu.pref("notifyJavaScriptLinks"))
				this.notify(
					this.getLocalized("title"),
					this.getLocalized("JavaScriptLink" + (loadJS ? "" : "Click")),
					true,
					loadJS ? null : loadJSFunc
				);
			loadJS && loadJSFunc();
			return;
		}
		else if(voidURI || this.isDummyURI(a, href, e)) {
			var evts = this.createMouseEvents(e, a, ["mousedown", "mouseup", "click"], 0);

			var _this = this;
			var loadVoidFunc = function() {
				_this.setJSLinksBackgroundPolicy();
				evts();
				_this.restoreJSLinksBackgroundPolicy();
			};

			var loadVoid = this.pu.pref("loadVoidLinksWithHandlers");
			if(this.pu.pref("notifyVoidLinksWithHandlers"))
				this.notify(
					this.getLocalized("title"),
					this.getLocalized("clickHandler" + (loadVoid ? "" : "Click")),
					true,
					loadVoid ? null : loadVoidFunc
				);
			loadVoid && loadVoidFunc();
			return;
		}

		var tbr = gBrowser;
		var flp = this.pu.pref("filesLinksPolicy");
		if(flp > 0) { // 0 - not check, 1 - load in current tab, 2 - show context menu
			var rePref = "filesLinksMask";
			var re = this.pu.pref(rePref);
			if(re) try {
				var _re = new RegExp(re, "i");
				if(_re.test(href)) {
					if(flp == 1)
						tbr.loadURI(href, this.getReferer());
					else
						this.showContextMenu();
					return;
				}
			}
			catch(e) {
				Components.utils.reportError(e);
				Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
					.getService(Components.interfaces.nsIPromptService)
					.alert(
						window,
						this.getLocalized("RegExpErrorTitle"),
						this.getLocalized("RegExpError")
							.replace("%r", re)
							.replace("%p", this.pu.prefNS + rePref)
							.replace("%e", e)
					);
			}
		}

		if("Highlander" in window) { // https://addons.mozilla.org/firefox/addon/4086
			var tab = Highlander.findTabForURI(makeURI(href));
			if(tab) {
				Highlander.selectTab(tab);
				return;
			}
		}

		// Open a new tab as a child of the current tab (Tree Style Tab)
		// http://piro.sakura.ne.jp/xul/_treestyletab.html.en#api
		if((this.itemType == "link" || this.itemType == "img") && "TreeStyleTabService" in window)
			TreeStyleTabService.readyToOpenChildTab(gBrowser.selectedTab);

		if("tabkit" in window) // Tab Kit https://addons.mozilla.org/firefox/addon/5447/
			tabkit.addingTab("related");

		var tab = tbr.addTab(href, this.getReferer());
		if(!this.pu.pref("loadInBackground"))
			tbr.selectedTab = tab;

		if("tabkit" in window)
			tabkit.addingTabOver();
	},
	contextmenuHandler: function(e) {
		if(this.enabled && this.stopContextMenu)
			this.stopEvent(e);
	},
	popupshowingHandler: function(e) {
		// Force prevent any popup
		// This is for RightToClick extension https://addons.mozilla.org/firefox/addon/righttoclick/
		if(this.enabled && this.stopContextMenu) {
			var ln = e.target.localName.toLowerCase();
			if(ln == "menupopup" || ln == "popup")
				this.stopEvent(e);
		}
	},
	cancelDelayedContext: function() {
		clearTimeout(this.showContextTimeout);
	},
	mousemoveHandler: function(e) {
		this.saveXY(e);
		if(!this.disallowMousemove)
			return;
		var mp = this.mousemoveParams, x = e.screenX, y = e.screenY;
		this.mousemoveParams = {
			dist: mp.dist + Math.sqrt(
				Math.pow(mp.screenX - x, 2) +
				Math.pow(mp.screenY - y, 2)
			),
			screenX: x,
			screenY: y
		};
		if(this.mousemoveParams.dist < this.pu.pref("disallowMousemoveDist"))
			return;

		this.runned = true;
		this.stopContextMenu = false;
		this.cancelDelayedContext();

		this.removeMoveHandler();
	},
	removeMoveHandler: function() {
		window.removeEventListener("mousemove", this, true);
		this.mousemoveParams = null;
	},
	_xy: null,
	saveXY: function(e) {
		this._xy = {
			screenX: e.screenX,
			screenY: e.screenY,
			clientX: e.clientX,
			clientY: e.clientY,
			__proto__: null
		};
	},

	loadJSLink: function() {
		var disId = "dom.disable_open_during_load";
		var dis = this.pu.getPref(disId, true);
		this.pu.setPref(disId, false); // allow window.open( ... )
		this.setJSLinksBackgroundPolicy();

		var a = this.item;
		var href = this.getHref(a);
		if(this.itemType == "bookmark" || this.itemType == "historyItem")
			gBrowser.loadURI(href); // bookmarklets
		else
			a.ownerDocument.location.href = href; // frames?

		this.restoreJSLinksBackgroundPolicy();
		if(this.fxVersion >= 3) // Bug? And dispatchEvent does not works also (fx block new window)...
			setTimeout(function(_this) {
				_this.pu.setPref(disId, dis);
			}, 5, this); // And now pop-up blocker does not work... =/
		else
			this.pu.setPref(disId, dis);
	},
	setJSLinksBackgroundPolicy: function() {
		var bgId = "browser.tabs.loadDivertedInBackground";
		this._origJSLinksInBg = this.pu.getPref(bgId, false);
		this.pu.setPref(bgId, this.pu.pref("loadJavaScriptLinksInBackground"));
	},
	restoreJSLinksBackgroundPolicy: function() {
		var origBg = this._origJSLinksInBg;
		setTimeout(function(_this) { // For Firefox 3.0+, timeout should be > 0 for Firefox 11.0+
			_this.pu.setPref("browser.tabs.loadDivertedInBackground", origBg);
		}, 5, this);
	},
	getReferer: function() {
		// Must be undefined for Firefox 3.6+ - see addTab() method in chrome://browser/content/tabbrowser.xml
		return (this.itemType == "link" || this.itemType == "img") && this.pu.pref("sendReferer")
			? makeURI(this.item.ownerDocument.location.href) // see chrome://global/content/contentAreaUtils.js
			: undefined;
	},
	showContextMenu: function() {
		this.stopContextMenu = false;
		this.createMouseEvents(
			this._xy || this.event,
			this.origItem || this.item,
			["mousedown", "mouseup", "contextmenu"], 2
		)();
	},
	createMouseEvents: function(origEvt, item, evtTypes, button) {
		var evts = evtTypes.map(function(evtType) {
			return this.createMouseEvent(origEvt, item, evtType, button);
		}, this);
		var _this = this;
		return function() {
			_this.enabled = false;
			evts.forEach(function(evt) {
				item.dispatchEvent(evt);
			});
			_this.enabled = true;
		};
	},
	createMouseEvent: function(origEvt, item, evtType, button) {
		item = item || origEvt.originalTarget;
		var doc = item.ownerDocument;
		var evt = doc.createEvent("MouseEvents");
		evt.initMouseEvent( // https://developer.mozilla.org/en/DOM/event.initMouseEvent
			evtType, true /* canBubble */, true /* cancelable */, doc.defaultView, 1,
			origEvt.screenX, origEvt.screenY, origEvt.clientX, origEvt.clientY,
			false, false, false, false,
			button, null
		);
		return evt;
	},

	// GUI:
	toggleStatus: function(notify) {
		this.enabled = !this.enabled;
		this.pu.pref("enabled", this.enabled);
		this.setStatus(this.enabled);
		if(
			!notify
			|| this.isElementVisible(this.status)
			|| this.isElementVisible(this.tbb)
		)
			return;
		this.notify(
			this.getLocalized("title"),
			this.getLocalized(this.enabled ? "enabled" : "disabled"),
			this.enabled
		);
	},
	isElementVisible: function(elt) {
		if(!elt)
			return false;
		var bo = elt.boxObject;
		return bo.height > 0 && bo.width > 0;
	},
	setStatus: function(enabled) {
		if(enabled === undefined)
			enabled = this.pu.pref("enabled");
		this.enabled = enabled;
		var st = this.status;
		var stVal = enabled ? "enabled" : "disabled";
		st.setAttribute("rl_status", stVal);
		var tt = this.getLocalized("title") + " " + this.getLocalized(stVal);
		st.tooltipText = tt;
		this.mi.setAttribute("checked", enabled);
		var miApp = this.miApp;
		miApp && miApp.setAttribute("checked", enabled);
		var tbb = this.tbb || this.paletteButton;
		if(!tbb)
			return;
		tbb.setAttribute("rl_status", stVal);
		tbb.tooltipText = tt;
		if(this.pu.pref("ui.toolbarbuttonCheckedStyle")) {
			tbb.setAttribute("type", "checkbox");
			tbb.setAttribute("checked", enabled);
		}
		else {
			tbb.removeAttribute("type");
			tbb.removeAttribute("checked");
		}
	},
	setItemsMode: function() {
		this.mi.hidden = !this.pu.pref("ui.showInToolsMenu");
		var miApp = this.miApp;
		if(miApp) {
			var sepApp = this.miAppSep;
			var hide = !this.pu.pref("ui.showInAppMenu");
			miApp.hidden = hide;
			sepApp.hidden = hide || !this.pu.pref("ui.showAppMenuSeparator");
		}
		this.status.hidden = !this.pu.pref("ui.showInStatusbar");
	},
	notify: function(ttl, txt, enabledImg, fnc) {
		var dur = this.pu.pref("notifyOpenTime");
		if(dur < 0)
			 return;
		if(enabledImg === undefined)
			enabledImg = true;
		window.openDialog(
			 "chrome://rightlinks/content/notify.xul",
			 "_blank",
			 "chrome,dialog=1,titlebar=0,popup=1",
			 dur, ttl, txt, enabledImg, fnc
		);
	},

	readPrefs: function(popup) {
		popup = popup || document.getElementById("rightLinks-settingsPopup");
		var closeMenu = this.pu.pref("ui.closeMenu") ? "auto" : "none";
		Array.forEach(
			popup.getElementsByAttribute("rl_pref", "*"),
			function(mi) {
				mi.setAttribute("closemenu", closeMenu);
				this.setPref(mi, true);
			},
			this
		);
		Array.forEach(
			popup.getElementsByAttribute("rl_depends", "*"),
			function(dep) {
				dep.setAttribute(
					"disabled",
					!Array.some(
						popup.getElementsByAttribute("rl_requiredFor", dep.getAttribute("rl_depends")),
						function(req) {
							return req.getAttribute("checked") == "true";
						}
					)
				);
			}
		);

		var miCheckedStyle = popup.getElementsByAttribute("rl_pref", "ui.toolbarbuttonCheckedStyle")[0];
		miCheckedStyle.hidden = miCheckedStyle.nextSibling.hidden = !this.isElementVisible(this.tbb);

		var miEnabled = popup.getElementsByAttribute("rl_pref", "enabled")[0];
		miEnabled.hidden = miEnabled.nextSibling.hidden = !("_showEnabledItem" in popup); // see options.xul

		if(!("_optionsCompatibilityChecked" in this)) {
			this._optionsCompatibilityChecked = true;
			// Attribute "closemenu" doesn't works
			popup.getElementsByAttribute("rl_pref", "ui.closeMenu")[0].hidden = this.fxVersion < 3;
			// Only in Firefox 4+
			popup.getElementsByAttribute("rl_pref", "ui.showInAppMenu")[0].hidden = !this.miApp;
		}
	},
	setPref: function(mi, readFlag) {
		var pName = mi.getAttribute("rl_pref");
		if(!pName)
			throw new Error("[Right Links]: Can't get pref name for \"" + mi.getAttribute("label") + "\"");
		if(readFlag) {
			mi.setAttribute("checked", !!this.pu.pref(pName));
			return;
		}
		this.pu.pref(pName, mi.getAttribute("checked") == "true");
		if(!this.pu.pref("ui.closeMenu") || pName == "ui.closeMenu")
			this.readPrefs();
	},
	prefsChanged: function(pName, pVal) {
		switch(pName) {
			case "ui.toolbarbuttonCheckedStyle":
			case "enabled":
				this.setStatus();
			break;
			case "ui.showInToolsMenu":
			case "ui.showInAppMenu":
			case "ui.showAppMenuSeparator":
			case "ui.showInStatusbar":
				this.setItemsMode();
		}
	},
	closePrefsMenu: function(e, popup) {
		if(e.button != 2 || !this.pu.pref("ui.closeMenuRightClick"))
			return;
		var pn = popup.triggerNode || document.popupNode; // https://bugzilla.mozilla.org/show_bug.cgi?id=383930
		if(!pn)
			return;
		this.closeMenus(pn);
		popup.hidePopup();
	},

	get strings() {
		delete this.strings;
		return this.strings = document.getElementById("rightLinks-strings");
	},
	getLocalized: function(id) {
		return this.strings.getString(id);
	},

	// Hotkeys:
	registerHotkeys: function() {
		this.pu.prefSvc.getBranch(this.pu.prefNS + "key.")
			.getChildList("", {})
			.forEach(this.registerHotkey, this);
	},
	registerHotkey: function(kId) {
		var keyStr = this.pu.pref("key." + kId);
		if(!keyStr) // Key is disabled
			return;
		var tokens = keyStr.split(" ");
		var key = tokens.pop() || " ";
		var modifiers = tokens.join(",");
		var kElt = document.getElementById("rightLinks-key-" + kId);
		kElt.removeAttribute("disabled");
		kElt.setAttribute(this.hasPrefix(key, "VK_") ? "keycode" : "key", key);
		kElt.setAttribute("modifiers", modifiers);
	}
};
window.addEventListener("load", rightLinks, false);